with
  ada.calendar,
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics.long_elementary_functions,
  terms,
  nlp;

procedure lights is

  use
    ada.calendar,
    text_io,
    ada.integer_text_io;

  subtype fmat3x1 is terms.fmat1 (1 .. 3);

  subtype fmat9x1 is terms.fmat1 (1 .. 9);

  subtype fmat2x6 is terms.fmat2 (1 .. 2, 1 .. 6);

  subtype fmat2x3 is terms.fmat2 (1 .. 2, 1 .. 3);

  type cablemat_rec is
    record
      young     : long_float;
      weight1   : long_float; -- force per unit length
      alpha     : long_float; -- thermal expansion coefficient
      area      : long_float;
      breakload : long_float;
      maxload   : long_float;
    end record;

  type cablemat_vec is array (integer range <>) of cablemat_rec;

  type cablemat_ptr is access cablemat_vec;

  type cable_rec is
    record
      mark    : boolean;
      mat     : integer;
      state   : integer range 0 .. 2 := 0;
      cut     : long_float := 0.0;
      tension : long_float := 0.0;
      glob    : terms.imat1 (1 .. 2);
      length  : long_float;
      strain0 : long_float;
      u       : fmat3x1;
    end record;

  type cable_vec is array (integer range <>) of cable_rec;

  type cable_ptr is access cable_vec;

  type membrmat_rec is
    record
      stiffness   : long_float; -- Young's modulus * thickness
      poisson     : long_float;
      weight2     : long_float; -- force per unit area
      breakstress : long_float; -- force per unit length
    end record;

  type membrmat_vec is array (integer range <>) of membrmat_rec;

  type membrmat_ptr is access membrmat_vec;

  type membr_rec is
    record
      mark     : boolean;
      mat      : integer;
      area     : long_float;
      l1       : long_float;
      l2       : long_float;
      l3       : long_float;
      glob     : terms.imat1 (1 .. 3);
      surface  : terms.fmat1 (1 .. 3);
      u1,u2,u3 : terms.fmat1 (1 .. 3);
      ht       : terms.fmat2 (1 .. 3, 1 .. 3);
    end record;

  type membr_vec is array (integer range <>) of membr_rec;

  type membr_ptr is access membr_vec;

  type framemat_rec is
    record
      young   : long_float;
      shear   : long_float; -- shear modulus
      weight3 : long_float; -- force per unit volume
    end record;

  type framemat_vec is array (integer range <>) of framemat_rec;

  type framemat_ptr is access framemat_vec;

  type frame_rec is
    record
      mat     : integer;
      release : terms.bmat1 (0 .. 12) := (0 .. 12 => false);
      group   : integer;
      glob    : terms.imat1 (1 .. 2);
      length  : long_float;
      ax      : long_float;
      ix      : long_float;
      iy      : long_float;
      iz      : long_float;
      stiff   : terms.fmat1 (0 .. 9);
      u,v,w   : fmat3x1;
    end record;

  type frame_vec is array (integer range <>) of frame_rec;

  type frame_ptr is access frame_vec;

  type spring_rec is
    record
      glob  : integer;
      axis  : integer;
      stiff : long_float;
    end record;

  type spring_vec is array (integer range <>) of spring_rec;

  type spring_ptr is access spring_vec;

  type origin_rec is
    record
      code    : integer range 0 .. 1;
      loading : integer;
      x       : terms.fmat1_ptr;
    end record;

  clock1,clock2           : time;
  halt                    : exception;
  optimal                 : boolean;
  nn,n,loading,maxlist    : integer;
  maxls,minor,maxminor,nl : integer;
  lstol,mntol,mnref       : long_float;
  maxnorm,f,csf,msf,temp  : long_float;
  cablemat                : cablemat_ptr;
  membrmat                : membrmat_ptr;
  framemat                : framemat_ptr;
  lastaxis                : terms.imat1_ptr;
  index                   : terms.imat2_ptr;
  xo,yo,zo,x,g            : terms.fmat1_ptr;
  known,force             : terms.fmat2_ptr;
  cable                   : cable_ptr;
  membr                   : membr_ptr;
  frame                   : frame_ptr;
  spring                  : spring_ptr;
  origin                  : origin_rec;
  io                      : file_type;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  procedure putchar (io : in file_type;
                     c  : in character;
                     n  : in integer) is

    use
      text_io;

  begin
    for k in 1 .. n loop
      put(io,c);
    end loop;
  end putchar;

  procedure getmaterial is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    nm,m : integer;
    io   : file_type;

  begin
    open(io, in_file, "material.txt");

    -- cable materials
    scan(io);
    get(io,nm);
    skip_line(io);

    cablemat := new cablemat_vec (1 .. nm);

    if nm > 0 then
      skip_line(io);

      for k in 1 .. nm loop
        get(io,m);
        get(io,cablemat(m).young);
        get(io,cablemat(m).weight1);
        get(io,cablemat(m).alpha);
        get(io,cablemat(m).area);
        get(io,cablemat(m).breakload);
        skip_line(io);
      end loop;
    end if;

    -- membrane materials
    scan(io);
    get(io,nm);
    skip_line(io);

    membrmat := new membrmat_vec (1 .. nm);

    if nm > 0 then
      skip_line(io);

      for k in 1 .. nm loop
        get(io,m);
        get(io,membrmat(m).stiffness);
        get(io,membrmat(m).poisson);
        get(io,membrmat(m).weight2);
        get(io,membrmat(m).breakstress);
        skip_line(io);
      end loop;
    end if;

    -- frame materials
    scan(io);
    get(io,nm);
    skip_line(io);

    framemat := new framemat_vec (1 .. nm);

    if nm > 0 then
      skip_line(io);

      for k in 1 .. nm loop
        get(io,m);
        get(io,framemat(m).young);
        get(io,framemat(m).shear);
        get(io,framemat(m).weight3);
        skip_line(io);
      end loop;
    end if;

    close(io);
  end getmaterial;

  procedure mov2glo (u,v,w : in fmat3x1;
                     l     : in fmat3x1;
                     g     : out fmat3x1) is

  begin
    -- local --> global
    g(1) := u(1) * l(1) + v(1) * l(2) + w(1) * l(3);
    g(2) := u(2) * l(1) + v(2) * l(2) + w(2) * l(3);
    g(3) := u(3) * l(1) + v(3) * l(2) + w(3) * l(3);
  end mov2glo;

  procedure mov2loc (u,v,w : in fmat3x1;
                     l     : out fmat3x1;
                     g     : in fmat3x1) is

  begin
    -- global --> local
    l(1) := u(1) * g(1) + u(2) * g(2) + u(3) * g(3);
    l(2) := v(1) * g(1) + v(2) * g(2) + v(3) * g(3);
    l(3) := w(1) * g(1) + w(2) * g(2) + w(3) * g(3);
  end mov2loc;

  procedure getaswl (frame : in frame_rec;
                     aswl  : out fmat2x6) is

    -- actions due to self weight on local reference

    aux : long_float;
    ae  : terms.fmat1 (1 .. 12);
    se  : terms.fmat2 (1 .. 12, 1 .. 12);

  begin
    -- actions due to self weight

    -- force
    aux := -framemat(frame.mat).weight3 * frame.ax * frame.length / 2.0;
    ae( 1) := frame.u(3) * aux;
    ae( 2) := frame.v(3) * aux;
    ae( 3) := frame.w(3) * aux;
    ae( 7) := ae( 1);
    ae( 8) := ae( 2);
    ae( 9) := ae( 3);

    -- moment
    aux := framemat(frame.mat).weight3 * frame.ax * frame.length ** 2 / 12.0;
    ae( 4) := 0.0;
    ae( 5) := (frame.u(1) * frame.v(2) - frame.u(2) * frame.v(1)) * aux;
    ae( 6) := (frame.u(1) * frame.w(2) - frame.u(2) * frame.w(1)) * aux;
    ae(10) := 0.0;
    ae(11) := -ae( 5);
    ae(12) := -ae( 6);

    if frame.release(0) then
      -- stiffness matrix
      for i in 1 .. 12 loop
        for j in 1 .. 12 loop
          se(i,j) := 0.0;
        end loop;
      end loop;

      se( 1, 1) :=  frame.stiff(0);
      se( 1, 7) := -frame.stiff(0);
      se( 2, 2) :=  frame.stiff(9);
      se( 2, 6) :=  frame.stiff(8);
      se( 2, 8) := -frame.stiff(9);
      se( 2,12) :=  frame.stiff(8);
      se( 3, 3) :=  frame.stiff(5);
      se( 3, 5) := -frame.stiff(4);
      se( 3, 9) := -frame.stiff(5);
      se( 3,11) := -frame.stiff(4);
      se( 4, 4) :=  frame.stiff(1);
      se( 4,10) := -frame.stiff(1);
      se( 5, 3) := -frame.stiff(4);
      se( 5, 5) :=  frame.stiff(3);
      se( 5, 9) :=  frame.stiff(4);
      se( 5,11) :=  frame.stiff(2);
      se( 6, 2) :=  frame.stiff(8);
      se( 6, 6) :=  frame.stiff(7);
      se( 6, 8) := -frame.stiff(8);
      se( 6,12) :=  frame.stiff(6);
      se( 7, 1) := -frame.stiff(0);
      se( 7, 7) :=  frame.stiff(0);
      se( 8, 2) := -frame.stiff(9);
      se( 8, 6) := -frame.stiff(8);
      se( 8, 8) :=  frame.stiff(9);
      se( 8,12) := -frame.stiff(8);
      se( 9, 3) := -frame.stiff(5);
      se( 9, 5) :=  frame.stiff(4);
      se( 9, 9) :=  frame.stiff(5);
      se( 9,11) :=  frame.stiff(4);
      se(10, 4) := -frame.stiff(1);
      se(10,10) :=  frame.stiff(1);
      se(11, 3) := -frame.stiff(4);
      se(11, 5) :=  frame.stiff(2);
      se(11, 9) :=  frame.stiff(4);
      se(11,11) :=  frame.stiff(3);
      se(12, 2) :=  frame.stiff(8);
      se(12, 6) :=  frame.stiff(6);
      se(12, 8) := -frame.stiff(8);
      se(12,12) :=  frame.stiff(7);

      -- release
      for k in 1 .. 12 loop
        if frame.release(k) then
          -- update actions
          aux := ae(k) / se(k,k);

          for i in 1 .. 12 loop
            if i /= k and se(i,k) /= 0.0 then
              ae(i) := ae(i) - se(i,k) * aux;
            end if;
          end loop;

          ae(k) := 0.0;

          -- update stiffness matrix
          for i in 1 .. 12 loop
            if i /= k and se(i,k) /= 0.0 then
              aux := se(i,k) / se(k,k);
              for j in 1 .. i loop
                if j /= k and se(k,j) /= 0.0 then
                  se(i,j) := se(i,j) - aux * se(k,j);
                end if;
              end loop;
            end if;
          end loop;

          for j in 1 .. k - 1 loop
            se(k,j) := 0.0;
          end loop;

          for i in k + 1 .. 12 loop
            se(i,k) := 0.0;
          end loop;

          se(k,k) := 0.0;

          -- simmetry
          for i in 1 .. 12 loop
            for j in 1 .. i - 1 loop
              se(j,i) := se(i,j);
            end loop;
          end loop;
        end if;
      end loop;
    end if;

    -- actions due to self weight
    aswl(1,1) := ae( 1);
    aswl(1,2) := ae( 2);
    aswl(1,3) := ae( 3);
    aswl(1,4) := ae( 4);
    aswl(1,5) := ae( 5);
    aswl(1,6) := ae( 6);

    aswl(2,1) := ae( 7);
    aswl(2,2) := ae( 8);
    aswl(2,3) := ae( 9);
    aswl(2,4) := ae(10);
    aswl(2,5) := ae(11);
    aswl(2,6) := ae(12);
  end getaswl;

  procedure getaswg (frame : in frame_rec;
                     aswg  : out fmat2x6) is

    -- actions due to self weight on global reference

    g,l : fmat3x1;

  begin
    -- local reference
    getaswl(frame,aswg);

    -- global reference
    for loc in 1 .. 2 loop
      -- force
      for axis in 1 .. 3 loop
        l(axis) := aswg(loc,axis);
      end loop;

      mov2glo(frame.u,frame.v,frame.w,l,g);

      for axis in 1 .. 3 loop
        aswg(loc,axis) := g(axis);
      end loop;

      -- moment
      for axis in 4 .. 6 loop
        l(axis - 3) := aswg(loc,axis);
      end loop;

      mov2glo(frame.u,frame.v,frame.w,l,g);

      for axis in 4 .. 6 loop
        aswg(loc,axis) := g(axis - 3);
      end loop;
    end loop;
  end getaswg;

  procedure cableobj (cable                   : in cable_rec;
                      epsilon,strain,axial,fe : out long_float;
                      x                       : in terms.fmat1_ptr;
                      ge                      : out fmat2x3) is

    use
      ada.numerics.long_elementary_functions;

    glob,i          : integer;
    c,l,alpha,normw : long_float;
    z,w             : terms.fmat1 (1 .. 3);
    xe              : terms.fmat2 (1 .. 2, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 2 loop
      glob := cable.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- calculate z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / cable.length;
    end loop;

    -- calculate w
    normw := 0.0;
    for axis in 1 .. 3 loop
      w(axis) := cable.u(axis) + z(axis);
      normw := normw + w(axis) ** 2;
    end loop;
    normw := sqrt(normw);

    -- calculate alpha
    alpha := 0.0;
    for axis in 1 .. 3 loop
      alpha := alpha + z(axis) * (cable.u(axis) + w(axis));
    end loop;

    -- calculate epsilon, strain, axial, fe
    epsilon := alpha / (normw + 1.0);

    case cable.state is
      -- normal state
      when 0 =>
      l := 1.0 + cable.strain0;
      strain := (epsilon - cable.strain0) / l;
      axial := cablemat(cable.mat).young * strain * cablemat(cable.mat).area;
      fe := 0.5 * axial * strain * cable.length * l;

      -- cut state
      when 1 =>
      c := cable.cut / cable.length;
      l := (1.0 - c) * (1.0 + cable.strain0);
      strain := (epsilon + c - cable.strain0 * (1.0 - c)) / l;
      axial := cablemat(cable.mat).young * strain * cablemat(cable.mat).area;
      fe := 0.5 * axial * strain * cable.length * l;

      -- tension state
      when 2 =>
      strain := cable.tension / cablemat(cable.mat).young;
      axial := cable.tension * cablemat(cable.mat).area;
      fe := cable.length * axial * normw;
    end case;

    -- global gradient
    c := axial / normw;
    for axis in 1 .. 3 loop
      alpha := c * w(axis);
      ge(1,axis) := -alpha;
      ge(2,axis) :=  alpha;
    end loop;
  end cableobj;

  procedure membrobj (membr          : in membr_rec;
                      st1,st2,st3,fe : out long_float;
                      x              : in terms.fmat1_ptr;
                      ge             : out fmat9x1) is

    glob,i               : integer;
    e1,e2,e3,sv1,sv2,sv3 : long_float;
    z                    : terms.fmat1 (1 .. 3);
    de1,de2,de3          : terms.fmat1 (1 .. 9);
    xe                   : terms.fmat2 (1 .. 3, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 3 loop
      glob := membr.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- e1
    for axis in 1 .. 3 loop
      z(axis) := (xe(3,axis) - xe(2,axis)) / membr.l1;
    end loop;

    e1 := 0.0;
    for axis in 1 .. 3 loop
      e1 := e1 + z(axis) * (membr.u1(axis) + 0.5 * z(axis));
    end loop;

    -- de1
    de1(1) := 0.0;
    de1(2) := 0.0;
    de1(3) := 0.0;

    de1(4) := -(membr.u1(1) + z(1)) / membr.l1;
    de1(5) := -(membr.u1(2) + z(2)) / membr.l1;
    de1(6) := -(membr.u1(3) + z(3)) / membr.l1;

    de1(7) := -de1(4);
    de1(8) := -de1(5);
    de1(9) := -de1(6);

    -- e2
    for axis in 1 .. 3 loop
      z(axis) := (xe(1,axis) - xe(3,axis)) / membr.l2;
    end loop;

    e2 := 0.0;
    for axis in 1 .. 3 loop
      e2 := e2 + z(axis) * (membr.u2(axis) + 0.5 * z(axis));
    end loop;

    -- de2
    de2(1) := (membr.u2(1) + z(1)) / membr.l2;
    de2(2) := (membr.u2(2) + z(2)) / membr.l2;
    de2(3) := (membr.u2(3) + z(3)) / membr.l2;

    de2(4) := 0.0;
    de2(5) := 0.0;
    de2(6) := 0.0;

    de2(7) := -de2(1);
    de2(8) := -de2(2);
    de2(9) := -de2(3);

    -- e3
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / membr.l3;
    end loop;

    e3 := 0.0;
    for axis in 1 .. 3 loop
      e3 := e3 + z(axis) * (membr.u3(axis) + 0.5 * z(axis));
    end loop;

    -- de3
    de3(1) := -(membr.u3(1) + z(1)) / membr.l3;
    de3(2) := -(membr.u3(2) + z(2)) / membr.l3;
    de3(3) := -(membr.u3(3) + z(3)) / membr.l3;

    de3(4) := -de3(1);
    de3(5) := -de3(2);
    de3(6) := -de3(3);

    de3(7) := 0.0;
    de3(8) := 0.0;
    de3(9) := 0.0;

    -- stress * thickness
    st1 := membr.ht(1,1) * e1 + membr.ht(1,2) * e2 + membr.ht(1,3) * e3;
    st2 := membr.ht(2,1) * e1 + membr.ht(2,2) * e2 + membr.ht(2,3) * e3;
    st3 := membr.ht(3,1) * e1 + membr.ht(3,2) * e2 + membr.ht(3,3) * e3;

    -- stress * volume
    sv1 := st1 * membr.area;
    sv2 := st2 * membr.area;
    sv3 := st3 * membr.area;

    -- global gradient
    for i in 1 .. 9 loop
      ge(i) := sv1 * de1(i) + sv2 * de2(i) + sv3 * de3(i);
    end loop;

    -- objective
    fe := (sv1 * e1 + sv2 * e2 + sv3 * e3) / 2.0;
  end membrobj;

  procedure frameobj (frame   : in frame_rec;
                      fe      : out long_float;
                      x       : in terms.fmat1_ptr;
                      gel,geg : out fmat2x6) is

    glob,i  : integer;
    aux,sum : long_float;
    g,l     : fmat3x1;
    de,ge   : terms.fmat1 (1 .. 12);
    xe      : terms.fmat2 (1 .. 2, 1 .. 6);
    se      : terms.fmat2 (1 .. 12, 1 .. 12);

  begin
    -- global displacements
    for loc in 1 .. 2 loop
      glob := frame.glob(loc);
      for axis in 1 .. 6 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- local displacements
    for loc in 1 .. 2 loop
      -- force
      for axis in 1 .. 3 loop
        g(axis) := xe(loc,axis);
      end loop;

      mov2loc(frame.u,frame.v,frame.w,l,g);

      for axis in 1 .. 3 loop
        xe(loc,axis) := l(axis);
      end loop;

      -- moment
      for axis in 4 .. 6 loop
        g(axis - 3) := xe(loc,axis);
      end loop;

      mov2loc(frame.u,frame.v,frame.w,l,g);

      for axis in 4 .. 6 loop
        xe(loc,axis) := l(axis - 3);
      end loop;
    end loop;

    -- local gradient
    if frame.release(0) then
      -- stiffness matrix
      for i in 1 .. 12 loop
        for j in 1 .. 12 loop
          se(i,j) := 0.0;
        end loop;
      end loop;

      se( 1, 1) :=  frame.stiff(0);
      se( 1, 7) := -frame.stiff(0);
      se( 2, 2) :=  frame.stiff(9);
      se( 2, 6) :=  frame.stiff(8);
      se( 2, 8) := -frame.stiff(9);
      se( 2,12) :=  frame.stiff(8);
      se( 3, 3) :=  frame.stiff(5);
      se( 3, 5) := -frame.stiff(4);
      se( 3, 9) := -frame.stiff(5);
      se( 3,11) := -frame.stiff(4);
      se( 4, 4) :=  frame.stiff(1);
      se( 4,10) := -frame.stiff(1);
      se( 5, 3) := -frame.stiff(4);
      se( 5, 5) :=  frame.stiff(3);
      se( 5, 9) :=  frame.stiff(4);
      se( 5,11) :=  frame.stiff(2);
      se( 6, 2) :=  frame.stiff(8);
      se( 6, 6) :=  frame.stiff(7);
      se( 6, 8) := -frame.stiff(8);
      se( 6,12) :=  frame.stiff(6);
      se( 7, 1) := -frame.stiff(0);
      se( 7, 7) :=  frame.stiff(0);
      se( 8, 2) := -frame.stiff(9);
      se( 8, 6) := -frame.stiff(8);
      se( 8, 8) :=  frame.stiff(9);
      se( 8,12) := -frame.stiff(8);
      se( 9, 3) := -frame.stiff(5);
      se( 9, 5) :=  frame.stiff(4);
      se( 9, 9) :=  frame.stiff(5);
      se( 9,11) :=  frame.stiff(4);
      se(10, 4) := -frame.stiff(1);
      se(10,10) :=  frame.stiff(1);
      se(11, 3) := -frame.stiff(4);
      se(11, 5) :=  frame.stiff(2);
      se(11, 9) :=  frame.stiff(4);
      se(11,11) :=  frame.stiff(3);
      se(12, 2) :=  frame.stiff(8);
      se(12, 6) :=  frame.stiff(6);
      se(12, 8) := -frame.stiff(8);
      se(12,12) :=  frame.stiff(7);

      -- update stiffness matrix
      for k in 1 .. 12 loop
        if frame.release(k) then
          for i in 1 .. 12 loop
            if i /= k and se(i,k) /= 0.0 then
              aux := se(i,k) / se(k,k);
              for j in 1 .. i loop
                if j /= k and se(k,j) /= 0.0 then
                  se(i,j) := se(i,j) - aux * se(k,j);
                end if;
              end loop;
            end if;
          end loop;

          for j in 1 .. k - 1 loop
            se(k,j) := 0.0;
          end loop;

          for i in k + 1 .. 12 loop
            se(i,k) := 0.0;
          end loop;

          se(k,k) := 0.0;

          -- simmetry
          for i in 1 .. 12 loop
            for j in 1 .. i - 1 loop
              se(j,i) := se(i,j);
            end loop;
          end loop;
        end if;
      end loop;

      -- local displacements
      de( 1) := xe(1,1);
      de( 2) := xe(1,2);
      de( 3) := xe(1,3);
      de( 4) := xe(1,4);
      de( 5) := xe(1,5);
      de( 6) := xe(1,6);

      de( 7) := xe(2,1);
      de( 8) := xe(2,2);
      de( 9) := xe(2,3);
      de(10) := xe(2,4);
      de(11) := xe(2,5);
      de(12) := xe(2,6);

      -- local gradient
      for i in 1 .. 12 loop
        sum := 0.0;

        for j in 1 .. 12 loop
          if se(i,j) /= 0.0 then
            sum := sum + se(i,j) * de(j);
          end if;
        end loop;

        ge(i) := sum;
      end loop;

      -- local gradient
      gel(1,1) := ge( 1);
      gel(1,2) := ge( 2);
      gel(1,3) := ge( 3);
      gel(1,4) := ge( 4);
      gel(1,5) := ge( 5);
      gel(1,6) := ge( 6);

      gel(2,1) := ge( 7);
      gel(2,2) := ge( 8);
      gel(2,3) := ge( 9);
      gel(2,4) := ge(10);
      gel(2,5) := ge(11);
      gel(2,6) := ge(12);
    else
      -- local gradient
      gel(1,1) := frame.stiff(0) * (xe(1,1) - xe(2,1));
      gel(1,2) := frame.stiff(9) * (xe(1,2) - xe(2,2)) + frame.stiff(8) * (xe(1,6) + xe(2,6));
      gel(1,3) := frame.stiff(5) * (xe(1,3) - xe(2,3)) - frame.stiff(4) * (xe(1,5) + xe(2,5));
      gel(1,4) := frame.stiff(1) * (xe(1,4) - xe(2,4));
      gel(1,5) := frame.stiff(4) * (xe(2,3) - xe(1,3)) + frame.stiff(3) * xe(1,5) + frame.stiff(2) * xe(2,5);
      gel(1,6) := frame.stiff(8) * (xe(1,2) - xe(2,2)) + frame.stiff(7) * xe(1,6) + frame.stiff(6) * xe(2,6);

      gel(2,1) := -gel(1,1);
      gel(2,2) := -gel(1,2);
      gel(2,3) := -gel(1,3);
      gel(2,4) := -gel(1,4);
      gel(2,5) := frame.stiff(4) * (xe(2,3) - xe(1,3)) + frame.stiff(2) * xe(1,5) + frame.stiff(3) * xe(2,5);
      gel(2,6) := frame.stiff(8) * (xe(1,2) - xe(2,2)) + frame.stiff(6) * xe(1,6) + frame.stiff(7) * xe(2,6);
    end if;

    -- objective
    fe := 0.0;
    for loc in 1 .. 2 loop
      for axis in 1 .. 6 loop
        fe := fe + xe(loc,axis) * gel(loc,axis);
      end loop;
    end loop;
    fe := 0.5 * fe;

    -- global gradient
    for loc in 1 .. 2 loop
      -- force
      for axis in 1 .. 3 loop
        l(axis) := gel(loc,axis);
      end loop;

      mov2glo(frame.u,frame.v,frame.w,l,g);

      for axis in 1 .. 3 loop
        geg(loc,axis) := g(axis);
      end loop;

      -- moment
      for axis in 4 .. 6 loop
        l(axis - 3) := gel(loc,axis);
      end loop;

      mov2glo(frame.u,frame.v,frame.w,l,g);

      for axis in 4 .. 6 loop
        geg(loc,axis) := g(axis - 3);
      end loop;
    end loop;
  end frameobj;

  procedure springobj (spring : in spring_rec;
                       fe,ge  : out long_float;
                       x      : in terms.fmat1_ptr) is

    i  : integer;
    xe : long_float;

  begin
    i := index(spring.glob,spring.axis);

    -- displacement
    if i > n then
      xe := known(spring.glob,spring.axis);
    else
      xe := x(i);
    end if;

    -- objective
    fe := 0.5 * spring.stiff * xe ** 2;

    -- gradient
    ge := spring.stiff * xe;
  end springobj;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

    glob,i,s                  : integer;
    epsilon,strain,axial,gseg : long_float;
    st1,st2,st3,action,fe     : long_float;
    gceg                      : fmat2x3;
    gmeg                      : fmat9x1;
    gfel,gfeg                 : fmat2x6;

  begin
    -- initiate objective, gradient
    f := 0.0;
    for glob in 1 .. nn loop
      for axis in 1 .. lastaxis(glob) loop
        i := index(glob,axis);
        if i <= n then
          action := force(glob,axis);
          f := f - action * x(i);
          g(i) := -action;
        end if;
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gceg);

      -- update objective
      f := f + fe;

      -- update gradient
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gceg(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;

    -- membrane elements
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gmeg);

      -- update objective
      f := f + fe;

      -- update gradient
      s := 0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gmeg(s);
          end if;
        end loop;
      end loop;
    end loop;

    -- frame elements
    for e in frame'range loop
      frameobj(frame(e),fe,x,gfel,gfeg);

      -- update objective
      f := f + fe;

      -- update gradient
      for loc in 1 .. 2 loop
        glob := frame(e).glob(loc);
        for axis in 1 .. 6 loop
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gfeg(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;

    -- spring elements
    for e in spring'range loop
      springobj(spring(e),fe,gseg,x);

      -- update objective
      f := f + fe;

      -- update gradient
      i := index(spring(e).glob,spring(e).axis);
      if i <= n then
        g(i) := g(i) + gseg;
      end if;
    end loop;
  end objective;

  package qn is new nlp.qnewton(objective);

  procedure getstructure is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    -- cosine of 0.6 degrees
    maxcos : constant long_float := 0.99995;

    e,glob,loc,axis     : integer;
    glob1,glob2,glob3   : integer;
    ne,nk,nr,s1,s2      : integer;
    ax,ix,iy,iz,length  : long_float;
    norm,young,shear    : long_float;
    sc0,sc1,sc2,sc3,sc4 : long_float;
    sc5,sc6,sc7,sc8,sc9 : long_float;
    cos1,cos2,cos3      : long_float;
    sin1,sin2,sin3      : long_float;
    sum,aux,stiff       : long_float;
    u,v,w               : terms.fmat1 (1 .. 3);
    a,b,c               : terms.fmat2 (1 .. 3, 1 .. 3);
    io                  : file_type;

  begin
    open(io, in_file, "structure.txt");

    -- solution parameters
    scan(io);
    get(io,lstol);
    get(io,maxls);
    get(io,mntol);
    get(io,mnref);
    get(io,maxminor);
    get(io,maxlist);
    skip_line(io);

    if mntol <= 0.0 then
      mntol := 0.001;
    end if;

    -- node coordinates
    scan(io);
    get(io,nn);
    skip_line(io);

    -- initiate xo, yo, zo
    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    skip_line(io);
    for k in 1 .. nn loop
      get(io,glob);
      get(io,xo(glob));
      get(io,yo(glob));
      get(io,zo(glob));
      skip_line(io);
    end loop;

    -- cable elements
    scan(io);
    get(io,ne);
    skip_line(io);

    -- initiate cable
    cable := new cable_vec (1 .. ne);

    if ne > 0 then
      skip_line(io);

      for k in 1 .. ne loop
        get(io,e);
        get(io,cable(e).glob(1));
        get(io,cable(e).glob(2));
        get(io,cable(e).mat);
        skip_line(io);

        glob1 := cable(e).glob(1);
        glob2 := cable(e).glob(2);

        u(1) := xo(glob2) - xo(glob1);
        u(2) := yo(glob2) - yo(glob1);
        u(3) := zo(glob2) - zo(glob1);
        norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);

        cable(e).length := norm;
        cable(e).u(1) := u(1) / norm;
        cable(e).u(2) := u(2) / norm;
        cable(e).u(3) := u(3) / norm;
      end loop;
    end if;

    -- membrane elements
    scan(io);
    get(io,ne);
    skip_line(io);

    -- initiate membr
    membr := new membr_vec (1 .. ne);

    if ne > 0 then
      skip_line(io);

      for k in 1 .. ne loop
        get(io,e);
        get(io,membr(e).glob(1));
        get(io,membr(e).glob(2));
        get(io,membr(e).glob(3));
        get(io,membr(e).mat);
        skip_line(io);

        glob1 := membr(e).glob(1);
        glob2 := membr(e).glob(2);
        glob3 := membr(e).glob(3);

        -- u
        membr(e).u1(1) := xo(glob3) - xo(glob2);
        membr(e).u1(2) := yo(glob3) - yo(glob2);
        membr(e).u1(3) := zo(glob3) - zo(glob2);

        membr(e).u2(1) := xo(glob1) - xo(glob3);
        membr(e).u2(2) := yo(glob1) - yo(glob3);
        membr(e).u2(3) := zo(glob1) - zo(glob3);

        membr(e).u3(1) := xo(glob2) - xo(glob1);
        membr(e).u3(2) := yo(glob2) - yo(glob1);
        membr(e).u3(3) := zo(glob2) - zo(glob1);

        -- u1
        norm := 0.0;
        for axis in 1 .. 3 loop
          norm := norm + membr(e).u1(axis) ** 2;
        end loop;
        norm := sqrt(norm);

        for axis in 1 .. 3 loop
          membr(e).u1(axis) := membr(e).u1(axis) / norm;
        end loop;

        membr(e).l1 := norm;

        -- u2
        norm := 0.0;
        for axis in 1 .. 3 loop
          norm := norm + membr(e).u2(axis) ** 2;
        end loop;
        norm := sqrt(norm);

        for axis in 1 .. 3 loop
          membr(e).u2(axis) := membr(e).u2(axis) / norm;
        end loop;

        membr(e).l2 := norm;

        -- u3
        norm := 0.0;
        for axis in 1 .. 3 loop
          norm := norm + membr(e).u3(axis) ** 2;
        end loop;
        norm := sqrt(norm);

        for axis in 1 .. 3 loop
          membr(e).u3(axis) := membr(e).u3(axis) / norm;
        end loop;

        membr(e).l3 := norm;

        -- cos ** 2
        cos1 := 0.0;
        for axis in 1 .. 3 loop
          cos1 := cos1 + membr(e).u2(axis) * membr(e).u3(axis);
        end loop;
        cos1 := cos1 ** 2;

        cos2 := 0.0;
        for axis in 1 .. 3 loop
          cos2 := cos2 + membr(e).u3(axis) * membr(e).u1(axis);
        end loop;
        cos2 := cos2 ** 2;

        cos3 := 0.0;
        for axis in 1 .. 3 loop
          cos3 := cos3 + membr(e).u1(axis) * membr(e).u2(axis);
        end loop;
        cos3 := cos3 ** 2;

        -- remove possible round off errors
        if cos1 > 1.0 then
          cos1 := 1.0;
        end if;

        if cos2 > 1.0 then
          cos2 := 1.0;
        end if;

        if cos3 > 1.0 then
          cos3 := 1.0;
        end if;

        -- sin ** 2
        sin1 := 1.0 - cos1;
        sin2 := 1.0 - cos2;
        sin3 := 1.0 - cos3;

        -- As a precaution against round off error, calculate area as
        -- the average of three values. These values should be exactly
        -- the same.

        sum := 0.0;
        sum := sum + membr(e).l2 * membr(e).l3 * sqrt(sin1);
        sum := sum + membr(e).l3 * membr(e).l1 * sqrt(sin2);
        sum := sum + membr(e).l1 * membr(e).l2 * sqrt(sin3);
        membr(e).area := sum / 6.0;

        -- matrix A (inverse)
        aux := 1.0 + 2.0 * cos1 * cos2 * cos3 - cos1 ** 2 - cos2 ** 2 - cos3 ** 2;
        a(1,1) := (1.0 - cos1 ** 2) / aux;
        a(2,2) := (1.0 - cos2 ** 2) / aux;
        a(3,3) := (1.0 - cos3 ** 2) / aux;
        a(1,2) := (cos1 * cos2 - cos3) / aux;
        a(1,3) := (cos1 * cos3 - cos2) / aux;
        a(2,3) := (cos2 * cos3 - cos1) / aux;
        a(2,1) := a(1,2);
        a(3,1) := a(1,3);
        a(3,2) := a(2,3);

        -- matrix B
        b(1,1) := 0.0;
        b(1,2) := sin3;
        b(1,3) := sin2;
        b(2,1) := sin3;
        b(2,2) := 0.0;
        b(2,3) := sin1;
        b(3,1) := sin2;
        b(3,2) := sin1;
        b(3,3) := 0.0;

        -- C <-- poisson * BA
        for i in 1 .. 3 loop
          for j in 1 .. 3 loop
            sum := 0.0;
            for k in 1 .. 3 loop
              sum := sum + b(i,k) * a(k,j);
            end loop;
            c(i,j) := membrmat(membr(e).mat).poisson * sum;
          end loop;
        end loop;

        -- C <-- C + I
        for k in 1 .. 3 loop
          c(k,k) := c(k,k) + 1.0;
        end loop;

        -- H * thickness
        aux := membrmat(membr(e).mat).stiffness / (1.0 - membrmat(membr(e).mat).poisson ** 2);
        for i in 1 .. 3 loop
          for j in 1 .. 3 loop
            sum := 0.0;
            for k in 1 .. 3 loop
              sum := sum + a(i,k) * c(k,j);
            end loop;
            membr(e).ht(i,j) := aux * sum;
          end loop;
        end loop;

        -- SURFACE is a vector with norm equal to the area of the
        -- element and with direction of the outward normal to the
        -- element.

        -- As a precaution against round off error, calculate surface
        -- vector as the average of three vectors. These vectors should
        -- be exactly the same.

        for axis in 1 .. 3 loop
          membr(e).surface(axis) := 0.0;
        end loop;

        for loc in 1 .. 3 loop
          glob1 := membr(e).glob(loc);
          glob2 := membr(e).glob(loc rem 3 + 1);
          glob3 := membr(e).glob((loc + 1) rem 3 + 1);

          -- u
          u(1) := xo(glob2) - xo(glob1);
          u(2) := yo(glob2) - yo(glob1);
          u(3) := zo(glob2) - zo(glob1);

          -- v
          v(1) := xo(glob3) - xo(glob1);
          v(2) := yo(glob3) - yo(glob1);
          v(3) := zo(glob3) - zo(glob1);

          -- w = u * v
          w(1) := u(2) * v(3) - v(2) * u(3);
          w(2) := u(3) * v(1) - v(3) * u(1);
          w(3) := v(2) * u(1) - v(1) * u(2);

          -- add w
          for axis in 1 .. 3 loop
            membr(e).surface(axis) := membr(e).surface(axis) + w(axis);
          end loop;
        end loop;

        for axis in 1 .. 3 loop
          membr(e).surface(axis) := membr(e).surface(axis) / 3.0;
        end loop;

        norm := 0.0;
        for axis in 1 .. 3 loop
          norm := norm + membr(e).surface(axis) ** 2;
        end loop;

        norm :=  membr(e).area / sqrt(norm);

        for axis in 1 .. 3 loop
          membr(e).surface(axis) := membr(e).surface(axis) * norm;
        end loop;
      end loop;
    end if;

    -- frame elements
    scan(io);
    get(io,ne);
    skip_line(io);

    -- initiate frame
    frame := new frame_vec (1 .. ne);

    if ne > 0 then
      skip_line(io);

      for k in 1 .. ne loop
        get(io,e);
        get(io,frame(e).glob(1));
        get(io,frame(e).glob(2));
        get(io,frame(e).mat);
        get(io,frame(e).group);
        skip_line(io);

        -- vector v
        get(io,frame(e).v(1));
        get(io,frame(e).v(2));
        get(io,frame(e).v(3));
        skip_line(io);

        get(io,frame(e).ax);
        get(io,frame(e).ix);
        get(io,frame(e).iy);
        get(io,frame(e).iz);
        skip_line(io);

        -- u
        glob1 := frame(e).glob(1);
        glob2 := frame(e).glob(2);
        u(1) := xo(glob2) - xo(glob1);
        u(2) := yo(glob2) - yo(glob1);
        u(3) := zo(glob2) - zo(glob1);

        -- normalize u
        norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
        u(1) := u(1) / norm;
        u(2) := u(2) / norm;
        u(3) := u(3) / norm;

        -- length
        frame(e).length := norm;

        -- v
        v(1) := frame(e).v(1);
        v(2) := frame(e).v(2);
        v(3) := frame(e).v(3);

        if v(1) = 0.0 and v(2) = 0.0 and v(3) = 0.0 then
          if abs u(3) > maxcos then
            v(1) := 0.0;
            v(2) := 1.0;
            v(3) := 0.0;
          else
            -- v = z * u
            v(1) := -u(2);
            v(2) := u(1);
            v(3) := 0.0;
          end if;
        end if;

        -- normalize v
        norm := sqrt(v(1) ** 2 + v(2) ** 2 + v(3) ** 2);
        v(1) := v(1) / norm;
        v(2) := v(2) / norm;
        v(3) := v(3) / norm;

        -- w = u * v
        w(1) := u(2) * v(3) - v(2) * u(3);
        w(2) := u(3) * v(1) - v(3) * u(1);
        w(3) := v(2) * u(1) - v(1) * u(2);

        -- normalize w
        norm := sqrt(w(1) ** 2 + w(2) ** 2 + w(3) ** 2);
        w(1) := w(1) / norm;
        w(2) := w(2) / norm;
        w(3) := w(3) / norm;

        -- u
        frame(e).u(1) := u(1);
        frame(e).u(2) := u(2);
        frame(e).u(3) := u(3);

        -- v
        frame(e).v(1) := v(1);
        frame(e).v(2) := v(2);
        frame(e).v(3) := v(3);

        -- w
        frame(e).w(1) := w(1);
        frame(e).w(2) := w(2);
        frame(e).w(3) := w(3);

        -- stiffness constants
        ax := frame(e).ax;
        ix := frame(e).ix;
        iy := frame(e).iy;
        iz := frame(e).iz;
        length := frame(e).length;
        young := framemat(frame(e).mat).young;
        shear := framemat(frame(e).mat).shear;

        sc0 := young * ax / length;
        sc1 := shear * ix / length;
        sc2 := 2.0 * young * iy / length;
        sc3 := 2.0 * sc2;
        sc4 := 3.0 * sc2 / length;
        sc5 := 2.0 * sc4 / length;
        sc6 := 2.0 * young * iz / length;
        sc7 := 2.0 * sc6;
        sc8 := 3.0 * sc6 / length;
        sc9 := 2.0 * sc8 / length;

        frame(e).stiff(0) := sc0;
        frame(e).stiff(1) := sc1;
        frame(e).stiff(2) := sc2;
        frame(e).stiff(3) := sc3;
        frame(e).stiff(4) := sc4;
        frame(e).stiff(5) := sc5;
        frame(e).stiff(6) := sc6;
        frame(e).stiff(7) := sc7;
        frame(e).stiff(8) := sc8;
        frame(e).stiff(9) := sc9;
      end loop;
    end if;

    -- frame releases
    scan(io);
    get(io,nr);
    skip_line(io);

    if nr > 0 then
      skip_line(io);

      for k in 1 .. nr loop
        get(io,e);
        get(io,loc);
        get(io,axis);
        skip_line(io);

        frame(e).release((loc - 1) * 6 + axis) := true;
        frame(e).release(0) := true;
      end loop;
    end if;

    -- initiate lastaxis
    lastaxis := new terms.imat1 (1 .. nn);

    for glob in 1 .. nn loop
      lastaxis(glob) := 0;
    end loop;

    for e in cable'range loop
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        if 3 > lastaxis(glob) then
          lastaxis(glob) := 3;
        end if;
      end loop;
    end loop;

    for e in membr'range loop
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        if 3 > lastaxis(glob) then
          lastaxis(glob) := 3;
        end if;
      end loop;
    end loop;

    for e in frame'range loop
      for loc in 1 .. 2 loop
        glob := frame(e).glob(loc);
        if 6 > lastaxis(glob) then
          lastaxis(glob) := 6;
        end if;
      end loop;
    end loop;

    -- initiate index
    index := new terms.imat2 (1 .. nn, 1 .. 6);

    for glob in 1 .. nn loop
      for axis in 1 .. 6 loop
        index(glob,axis) := 0;
      end loop;
    end loop;

    -- spring elements
    scan(io);
    get(io,ne);
    skip_line(io);

    spring := new spring_vec (1 .. ne);

    if ne > 0 then
      skip_line(io);

      for k in 1 .. ne loop
        get(io,e);
        get(io,glob);
        get(io,axis);
        get(io,stiff);
        skip_line(io);

        if axis > lastaxis(glob) then
          put("Displacement not allowed");
          new_line;
          put("Node : ");
          put(glob,5);
          put(" , Axis : ");
          put(axis,5);
          new_line;
          raise halt;
        end if;

        if index(glob,axis) = 1 then
          put("Spring element already declared");
          new_line;
          put("Node : ");
          put(glob,5);
          put(" , Axis : ");
          put(axis,5);
          new_line;
          raise halt;
        end if;

        spring(e).glob := glob;
        spring(e).axis := axis;
        spring(e).stiff := stiff;

        index(glob,axis) := 1;
      end loop;
    end if;

    -- known displacements
    known := new terms.fmat2 (1 .. nn, 1 .. 6);

    for glob in 1 .. nn loop
      for axis in 1 .. 6 loop
        known(glob,axis) := 0.0;
      end loop;
    end loop;

    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);

      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,known(glob,axis));
        skip_line(io);

        if axis > lastaxis(glob) then
          put("Displacement not allowed");
          new_line;
          put("Node : ");
          put(glob,5);
          put(" , Axis : ");
          put(axis,5);
          new_line;
          raise halt;
        end if;

        if index(glob,axis) = 2 then
          put("Known displacement already declared");
          new_line;
          put("Node : ");
          put(glob,5);
          put(" , Axis : ");
          put(axis,5);
          new_line;
          raise halt;
        end if;

        index(glob,axis) := 2;
      end loop;
    end if;

    close(io);

    -- initiate force
    force := new terms.fmat2 (1 .. nn, 1 .. 6);

    -- number of unknowns
    n := -nk;
    for glob in 1 .. nn loop
      for axis in 1 .. lastaxis(glob) loop
        n := n + 1;
      end loop;
    end loop;

    -- new sequence
    s1 := 0;
    s2 := n;
    for glob in 1 .. nn loop
      for axis in 1 .. lastaxis(glob) loop
        if index(glob,axis) = 2 then
          s2 := s2 + 1;
          index(glob,axis) := s2;
        else
          s1 := s1 + 1;
          index(glob,axis) := s1;
        end if;
      end loop;
    end loop;

    -- initiate x
    x := new terms.fmat1 (1 .. n);

    -- initiate g
    g := new terms.fmat1 (1 .. n);

    -- initiate origin
    origin.code := 0;
    origin.loading := 0;
    origin.x := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      origin.x(k) := 0.0;
    end loop;

    create(io, out_file, "cable.txt");
    close(io);

    create(io, out_file, "membrane.txt");
    close(io);

    create(io, out_file, "reaction.txt");
    close(io);

    create(io, out_file, "displacement.txt");
    close(io);

    create(io, out_file, "coordinate.txt");
    close(io);

    create(io, out_file, "frame.txt");
    close(io);

    create(io, out_file, "spring.txt");
    close(io);
  end getstructure;

  procedure getloading (io      : in file_type;
                        loading : out integer;
                        x       : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    code              : integer range 0 .. 4;
    e,glob,axis,na,nc : integer;
    action            : long_float;
    a                 : terms.fmat1 (1 .. 3);
    aswg              : fmat2x6;

  begin
    -- initiate force
    for glob in 1 .. nn loop
      for axis in 1 .. 6 loop
        force(glob,axis) := 0.0;
      end loop;
    end loop;

    -- self weight for cable elements
    for e in cable'range loop
      action := -cablemat(cable(e).mat).weight1 * cable(e).length / 2.0;

      glob := cable(e).glob(1);
      force(glob,3) := force(glob,3) + action;

      glob := cable(e).glob(2);
      force(glob,3) := force(glob,3) + action;
    end loop;

    -- self weight for membrane elements
    for e in membr'range loop
      action := -membrmat(membr(e).mat).weight2 * membr(e).area / 3.0;

      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        force(glob,3) := force(glob,3) + action;
      end loop;
    end loop;

    -- self weight for frame elements
    for e in frame'range loop
      getaswg(frame(e),aswg);

      for loc in 1 .. 2 loop
        glob := frame(e).glob(loc);
        for axis in 1 .. 6 loop
          force(glob,axis) := force(glob,axis) + aswg(loc,axis);
        end loop;
      end loop;
    end loop;

    -- loading parameters
    scan(io);
    get(io,loading);
    get(io,csf);
    get(io,msf);
    get(io,temp);
    get(io,origin.code);
    skip_line(io);

    -- nodal forces
    scan(io);
    get(io,na);
    skip_line(io);

    if na > 0 then
      skip_line(io);

      for k in 1 .. na loop
        get(io,glob);
        get(io,axis);
        get(io,action);
        skip_line(io);

        force(glob,axis) := force(glob,axis) + action;
      end loop;
    end if;

    -- change cable element state
    scan(io);
    get(io,nc);
    skip_line(io);

    if nc > 0 then
      skip_line(io);

      for k in 1 .. nc loop
        get(io,e);
        get(io,code);

        case code is
          -- set normal state
          when 0 =>
          cable(e).state := 0;
          cable(e).cut := 0.0;
          cable(e).tension := 0.0;

          -- set cut state with given cut value
          when 1 =>
          get(io,cable(e).cut);
          cable(e).state := 1;

          -- set tension state with given tension value
          when 2 =>
          get(io,cable(e).tension);
          cable(e).state := 2;

          -- set cut state with cut value given by current tension state
          when 3 =>
          cable(e).state := 1;

          -- set tension state with tension value given by current cut state
          when 4 =>
          cable(e).state := 2;
        end case;

        skip_line(io);
      end loop;
    end if;

    -- force per unit area in the Z direction on membrane elements
    scan(io);
    get(io,na);
    skip_line(io);

    if na > 0 then
      skip_line(io);

      for k in 1 .. na loop
        get(io,e);
        get(io,action);
        skip_line(io);

        action := action * membr(e).area / 3.0;

        for loc in 1 .. 3 loop
          glob := membr(e).glob(loc);
          force(glob,3) := force(glob,3) + action;
        end loop;
      end loop;
    end if;

    -- force per unit area orthogonal to the surface on membrane elements
    scan(io);
    get(io,na);
    skip_line(io);

    if na > 0 then
      skip_line(io);

      for k in 1 .. na loop
        get(io,e);
        get(io,action);
        skip_line(io);

        action := action / 3.0;

        for axis in 1 .. 3 loop
          a(axis) := action * membr(e).surface(axis);
        end loop;

        for loc in 1 .. 3 loop
          glob := membr(e).glob(loc);
          force(glob,1) := force(glob,1) + a(1);
          force(glob,2) := force(glob,2) + a(2);
          force(glob,3) := force(glob,3) + a(3);
        end loop;
      end loop;
    end if;

    -- set temperature strain for cable elements
    for e in cable'range loop
      cable(e).strain0 := temp * cablemat(cable(e).mat).alpha;
    end loop;

    -- initiate x
    for k in 1 .. n loop
      x(k) := 0.0;
    end loop;

    -- maxnorm
    if mnref <= 0.0 then
      objective(n,f,x,g);
      maxnorm := mntol * nlp.norm(n,g);
    else
      maxnorm := mntol * mnref;
    end if;
  end getloading;

  procedure finish (x : in terms.fmat1_ptr) is

    epsilon,strain,axial,fe : long_float;
    strain0,cut             : long_float;
    gceg                    : fmat2x3;

  begin
    -- equivalence between constant cut and constant tension
    for e in cable'range loop
      case cable(e).state is
        -- normal state
        when 0 =>
        null;

        -- cut state
        -- set tension with value given by current cut
        when 1 =>
        cableobj(cable(e),epsilon,strain,axial,fe,x,gceg);
        cable(e).tension := strain * cablemat(cable(e).mat).young;

        -- tension state
        -- set cut with value given by current tension
        when 2 =>
        cableobj(cable(e),epsilon,strain,axial,fe,x,gceg);
        strain0 := cable(e).strain0;
        cut := strain + strain0 + strain0 * strain - epsilon;
        cable(e).cut := cut * cable(e).length / (1.0 + strain0) / (1.0 + strain);
      end case;
    end loop;
  end finish;

  procedure putdispl (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    infinity : constant long_float := 1.0E+20;

    type bound_rec is
      record
        glob  : integer;
        displ : long_float;
      end record;

    i        : integer;
    displ    : long_float;
    min,max  : array (1 .. 3) of bound_rec;
    io       : file_type;

  begin
    open(io, append_file, "displacement.txt");

    new_line(io);
    new_line(io);
    new_line(io);
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Loading                                               = ");
    put(io,loading,5);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- displacements
    new_line(io);
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    put(io," Node             Displ X             Displ Y             Displ Z");
    new_line(io);

    put(io,"                    Rot X               Rot Y               Rot Z");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          putchar(io,' ',5);
          put(io,known(glob,axis),2,8,3);
        else
          putchar(io,' ',5);
          put(io,x(i),2,8,3);
        end if;
      end loop;
      new_line(io);

      if lastaxis(glob) > 3 then
        putchar(io,' ',5);
        for axis in 4 .. lastaxis(glob) loop
          i := index(glob,axis);
          if i > n then
            putchar(io,' ',5);
            put(io,known(glob,axis),2,8,3);
          else
            putchar(io,' ',5);
            put(io,x(i),2,8,3);
          end if;
        end loop;
        new_line(io);
      end if;
    end loop;

    -- minimum and maximum displacements
    for axis in 1 .. 3 loop
      min(axis).displ :=  infinity;
      min(axis).glob := 0;
      max(axis).displ := -infinity;
      max(axis).glob := 0;

      for glob in 1 .. nn loop
        i := index(glob,axis);

        if i <= n then
          displ := x(i) - origin.x(i);

          -- minimum displacements
          if displ < min(axis).displ then
            min(axis).displ := displ;
            min(axis).glob := glob;
          end if;

          -- maximum displacements
          if displ > max(axis).displ then
            max(axis).displ := displ;
            max(axis).glob := glob;
          end if;
        end if;
      end loop;
    end loop;

    new_line(io);
    new_line(io);

    putchar(io,'-',35);
    new_line(io);

    put(io,"| Origin from Loading     = ");
    put(io,origin.loading,5);
    put(io," |");
    new_line(io);

    putchar(io,'-',35);
    new_line(io);

    -- minimum displacements
    new_line(io);
    new_line(io);

    putchar(io,'-',35);
    new_line(io);

    put(io," Axis           Min Displ      Node");
    new_line(io);

    putchar(io,'-',35);
    new_line(io);

    new_line(io);

    for axis in 1 .. 3 loop
      if min(axis).glob > 0 then
        case axis is
          when 1 =>
          put(io,"    X");

          when 2 =>
          put(io,"    Y");

          when 3 =>
          put(io,"    Z");

          when others =>
          null;
        end case;

        putchar(io,' ',5);
        put(io,min(axis).displ,2,8,3);

        putchar(io,' ',5);
        put(io,min(axis).glob,5);

        new_line(io);
      end if;
    end loop;

    -- maximum displacements
    new_line(io);
    new_line(io);

    putchar(io,'-',35);
    new_line(io);

    put(io," Axis           Max Displ      Node");
    new_line(io);

    putchar(io,'-',35);
    new_line(io);

    new_line(io);

    for axis in 1 .. 3 loop
      if max(axis).glob > 0 then
        case axis is
          when 1 =>
          put(io,"    X");

          when 2 =>
          put(io,"    Y");

          when 3 =>
          put(io,"    Z");

          when others =>
          null;
        end case;

        putchar(io,' ',5);
        put(io,max(axis).displ,2,8,3);

        putchar(io,' ',5);
        put(io,max(axis).glob,5);

        new_line(io);
      end if;
    end loop;

    close(io);

    -- update origin
    if origin.code = 1 then
      origin.code := 0;
      origin.loading := loading;
      for k in 1 .. n loop
        origin.x(k) := x(k);
      end loop;
    end if;
  end putdispl;

  procedure vertex (glob : in integer;
                    v    : out fmat3x1) is

    i : integer;

  begin
    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i > n then
        v(axis) := known(glob,axis);
      else
        v(axis) := x(i);
      end if;
    end loop;

    v(1) := v(1) + xo(glob);
    v(2) := v(2) + yo(glob);
    v(3) := v(3) + zo(glob);
  end vertex;

  procedure putcoord (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    v  : fmat3x1;
    io : file_type;

  begin
    open(io, append_file, "coordinate.txt");

    new_line(io);
    new_line(io);
    new_line(io);
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Loading                                               = ");
    put(io,loading,5);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- coordinates
    new_line(io);
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    put(io," Node             Coord X             Coord Y             Coord Z");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,v);
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        put(io,v(axis),2,8,3);
      end loop;
      new_line(io);
    end loop;

    close(io);
  end putcoord;

  procedure putcable (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    i                       : integer;
    epsilon,strain,axial,fe : long_float;
    lrf                     : long_float;
    gceg                    : fmat2x3;
    io                      : file_type;

  begin
    open(io, append_file, "cable.txt");

    new_line(io);
    new_line(io);
    new_line(io);
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    put(io,"| Loading                             = ");
    put(io,loading,5);
    put(io," |");
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    new_line(io);
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    put(io,"Element                Load     Load/Resistance");
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    new_line(io);

    -- initiate maxload
    for e in cablemat'range loop
      cablemat(e).maxload := 0.0;
    end loop;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gceg);

      put(io,e,7);
      putchar(io,' ',5);
      put(io,axial,2,8,3);

      cable(e).mark := false;

      if axial < 0.0 then
        cable(e).mark := true;
        put(io,"         Compression");
      else
        lrf := axial / cablemat(cable(e).mat).breakload * csf;
        if lrf > 1.0 then
          putchar(io,' ',5);
          put(io,lrf,2,8,3);
        end if;
      end if;

      new_line(io);

      -- update maxload
      i := cable(e).mat;
      if axial > cablemat(i).maxload then
        cablemat(i).maxload := axial;
      end if;
    end loop;

    -- maximum load
    new_line(io);
    new_line(io);

    putchar(io,'-',25);
    new_line(io);

    put(io,"Cable            Max Load");
    new_line(io);

    putchar(io,'-',25);
    new_line(io);

    new_line(io);

    for e in cablemat'range loop
      axial := cablemat(e).maxload;
      if axial > 0.0 then
        put(io,e,5);
        putchar(io,' ',5);
        put(io,axial,2,8,3);
        new_line(io);
      end if;
    end loop;

    close(io);
  end putcable;

  procedure putmembr (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    aux,st1,st2,st3,fe,lrf : long_float;
    s1,s2,cos1,cos2,cos3   : long_float;
    gmeg                   : fmat9x1;
    io                     : file_type;

  begin
    open(io, append_file, "membrane.txt");

    new_line(io);
    new_line(io);
    new_line(io);
    new_line(io);

    putchar(io,'-',67);
    new_line(io);

    put(io,"| Loading                                                 = ");
    put(io,loading,5);
    put(io," |");
    new_line(io);

    putchar(io,'-',67);
    new_line(io);

    -- principal stresses
    new_line(io);
    new_line(io);

    putchar(io,'-',67);
    new_line(io);

    put(io,"Element            Stress-1            Stress-2     Load/Resistance");
    new_line(io);

    putchar(io,'-',67);
    new_line(io);

    new_line(io);

    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gmeg);

      -- cos ** 2
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 + membr(e).u2(axis) * membr(e).u3(axis);
      end loop;
      cos1 := cos1 ** 2;

      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 + membr(e).u3(axis) * membr(e).u1(axis);
      end loop;
      cos2 := cos2 ** 2;

      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 + membr(e).u1(axis) * membr(e).u2(axis);
      end loop;
      cos3 := cos3 ** 2;

      -- remove possible round off errors
      if cos1 > 1.0 then
        cos1 := 1.0;
      end if;

      if cos2 > 1.0 then
        cos2 := 1.0;
      end if;

      if cos3 > 1.0 then
        cos3 := 1.0;
      end if;

      -- principal stresses
      aux := 0.0;
      aux := aux + st1 * st2 * (cos3 - 0.5);
      aux := aux + st1 * st3 * (cos2 - 0.5);
      aux := aux + st2 * st3 * (cos1 - 0.5);
      aux := 4.0 * aux + st1 ** 2 + st2 ** 2 + st3 ** 2;
      aux := sqrt(aux);

      s1 := ((st1 + st2 + st3) + aux) / 2.0;
      s2 := ((st1 + st2 + st3) - aux) / 2.0;

      put(io,e,7);
      putchar(io,' ',5);
      put(io,s1,2,8,3);
      putchar(io,' ',5);
      put(io,s2,2,8,3);

      membr(e).mark := false;

      if s1 < 0.0 or s2 < 0.0 then
        membr(e).mark := true;
        put(io,"         Compression");
      else
        lrf := s1 / membrmat(membr(e).mat).breakstress * msf;
        if lrf > 1.0 then
          putchar(io,' ',5);
          put(io,lrf,2,8,3);
        end if;
      end if;

      new_line(io);
    end loop;

    close(io);
  end putmembr;

  procedure putframe (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    type bound_rec is
      record
        element : integer;
        value   : long_float;
      end record;

    g,ming,maxg : integer;
    fe,value    : long_float;
    gfel,gfeg,a : fmat2x6;
    bound       : array (1 .. 9) of bound_rec;
    io          : file_type;

  begin
    -- group numbering range
    ming := frame'last + 1;
    maxg := 0;

    for e in frame'range loop
      g := frame(e).group;

      if g < ming then
        ming := g;
      end if;

      if g > maxg then
        maxg := g;
      end if;
    end loop;

    open(io, append_file, "frame.txt");

    new_line(io);
    new_line(io);

    for g in ming .. maxg loop
      new_line(io);
      new_line(io);

      putchar(io,'-',75);
      new_line(io);

      put(io,"| Loading = ");
      put(io,loading,5);
      put(io," |");
      put(io,"                                       ");
      put(io,"| Group = ");
      put(io,g,5);
      put(io," |");
      new_line(io);

      putchar(io,'-',75);
      new_line(io);

      -- frame elements
      new_line(io);
      new_line(io);

      putchar(io,'-',75);
      new_line(io);

      put(io,"Element    Node             Force X             Force Y             Force Z");
      new_line(io);

      put(io,"                           Moment X            Moment Y            Moment Z");
      new_line(io);

      putchar(io,'-',75);
      new_line(io);

      -- initiate bound
      for s in 1 .. 9 loop
        bound(s).element := 0;
        bound(s).value := 0.0;
      end loop;

      for e in frame'range loop
        if frame(e).group = g then
          frameobj(frame(e),fe,x,gfel,gfeg);

          getaswl(frame(e),a);

          for loc in 1 .. 2 loop
            for axis in 1 .. 6 loop
              a(loc,axis) := gfel(loc,axis) - a(loc,axis);
            end loop;
          end loop;

          new_line(io);

          for loc in 1 .. 2 loop
            put(io,e,7);
            put(io,frame(e).glob(loc),8);

            for axis in 1 .. 3 loop
              putchar(io,' ',5);
              put(io,a(loc,axis),2,8,3);
            end loop;
            new_line(io);

            putchar(io,' ',15);

            for axis in 4 .. 6 loop
              putchar(io,' ',5);
              put(io,a(loc,axis),2,8,3);
            end loop;
            new_line(io);
          end loop;

          -- compression bound
          if a(1,1) > -a(2,1) then
            value := a(1,1);
          else
            value := -a(2,1);
          end if;

          if value > bound(1).value then
            bound(1).element := e;
            bound(1).value := value;
          end if;

          -- tension bound
          if -a(1,1) > a(2,1) then
            value := -a(1,1);
          else
            value := a(2,1);
          end if;

          if value > bound(2).value then
            bound(2).element := e;
            bound(2).value := value;
          end if;

          -- y flexure bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if abs a(loc,5) > value then
              value := abs a(loc,5);
            end if;
          end loop;

          if value > bound(3).value then
            bound(3).element := e;
            bound(3).value := value;
          end if;

          -- z flexure bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if abs a(loc,6) > value then
              value := abs a(loc,6);
            end if;
          end loop;

          if value > bound(4).value then
            bound(4).element := e;
            bound(4).value := value;
          end if;

          -- r flexure bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if sqrt(a(loc,5) ** 2 + a(loc,6) ** 2) > value then
              value := sqrt(a(loc,5) ** 2 + a(loc,6) ** 2);
            end if;
          end loop;

          if value > bound(5).value then
            bound(5).element := e;
            bound(5).value := value;
          end if;

          -- y shear bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if abs a(loc,2) > value then
              value := abs a(loc,2);
            end if;
          end loop;

          if value > bound(6).value then
            bound(6).element := e;
            bound(6).value := value;
          end if;

          -- z shear bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if abs a(loc,3) > value then
              value := abs a(loc,3);
            end if;
          end loop;

          if value > bound(7).value then
            bound(7).element := e;
            bound(7).value := value;
          end if;

          -- r shear bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if sqrt(a(loc,2) ** 2 + a(loc,3) ** 2) > value then
              value := sqrt(a(loc,2) ** 2 + a(loc,3) ** 2);
            end if;
          end loop;

          if value > bound(8).value then
            bound(8).element := e;
            bound(8).value := value;
          end if;

          -- torsion bound
          value := 0.0;
          for loc in 1 .. 2 loop
            if abs a(loc,4) > value then
              value := abs a(loc,4);
            end if;
          end loop;

          if value > bound(9).value then
            bound(9).element := e;
            bound(9).value := value;
          end if;
        end if;
      end loop;

      -- bound
      new_line(io);
      new_line(io);

      putchar(io,'-',53);
      new_line(io);

      put(io,"Max Action                Element               Value");
      new_line(io);

      putchar(io,'-',53);
      new_line(io);

      new_line(io);

      for s in 1 .. 9 loop
        if bound(s).element > 0 then
          case s is
            when 1 =>
            put(io,"Compression            ");

            when 2 =>
            put(io,"Tension                ");

            when 3 =>
            put(io,"Flexure Y              ");

            when 4 =>
            put(io,"Flexure Z              ");

            when 5 =>
            put(io,"Flexure (Y^2 + Z^2)^0.5");

            when 6 =>
            put(io,"Shear Y                ");

            when 7 =>
            put(io,"Shear Z                ");

            when 8 =>
            put(io,"Shear (Y^2 + Z^2)^0.5  ");

            when 9 =>
            put(io,"Torsion                ");

            when others =>
            null;
          end case;

          putchar(io,' ',5);
          put(io,bound(s).element,5);
          putchar(io,' ',5);
          put(io,bound(s).value,2,8,3);
          new_line(io);
        end if;
      end loop;
    end loop;

    close(io);
  end putframe;

  procedure putspring (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    fe,gseg : long_float;
    io      : file_type;

  begin
    open(io, append_file, "spring.txt");

    new_line(io);
    new_line(io);
    new_line(io);
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    put(io,"| Loading                   | ");
    put(io,loading,15);
    put(io," |");
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    new_line(io);
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    put(io,"Element      Node      Axis               Force");
    new_line(io);

    putchar(io,'-',47);
    new_line(io);

    new_line(io);

    -- spring elements
    for e in spring'range loop
      springobj(spring(e),fe,gseg,x);

      put(io,e,7);
      putchar(io,' ',5);
      put(io,spring(e).glob,5);
      putchar(io,' ',5);
      put(io,spring(e).axis,5);
      putchar(io,' ',5);
      put(io,gseg,2,8,3);
      new_line(io);
    end loop;

    close(io);
  end putspring;

  procedure putreact (minor     : in integer;
                      f,maxnorm : in long_float;
                      x         : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    glob,axis,last1,last2,s : integer;
    epsilon,strain,axial,fe : long_float;
    action,maxr1,maxr2      : long_float;
    st1,st2,st3,gseg        : long_float;
    gceg                    : fmat2x3;
    gmeg                    : fmat9x1;
    gfel,gfeg               : fmat2x6;
    io                      : file_type;

  begin
    open(io, append_file, "reaction.txt");

    -- initiate support reactions
    for glob in 1 .. nn loop
      for axis in 1 .. 6 loop
        force(glob,axis) := -force(glob,axis);
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),epsilon,strain,axial,fe,x,gceg);

      -- update support reactions
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          force(glob,axis) := force(glob,axis) + gceg(loc,axis);
        end loop;
      end loop;
    end loop;

    -- membrane elements
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gmeg);

      -- update support reactions
      s := 0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          force(glob,axis) := force(glob,axis) + gmeg(s);
        end loop;
      end loop;
    end loop;

    -- frame elements
    for e in frame'range loop
      frameobj(frame(e),fe,x,gfel,gfeg);

      -- update support reactions
      for loc in 1 .. 2 loop
        glob := frame(e).glob(loc);
        for axis in 1 .. 6 loop
          force(glob,axis) := force(glob,axis) + gfeg(loc,axis);
        end loop;
      end loop;
    end loop;

    -- spring elements
    for e in spring'range loop
      springobj(spring(e),fe,gseg,x);

      -- update support reactions
      glob := spring(e).glob;
      axis := spring(e).axis;
      force(glob,axis) := force(glob,axis) + gseg;
    end loop;

    -- maximum residue
    maxr1 := 0.0;
    maxr2 := 0.0;

    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        action := abs force(glob,axis);
        if index(glob,axis) <= n and action > maxr1 then
          maxr1 := action;
        end if;
      end loop;

      for axis in 4 .. lastaxis(glob) loop
        action := abs force(glob,axis);
        if index(glob,axis) <= n and action > maxr2 then
          maxr2 := action;
        end if;
      end loop;
    end loop;

    new_line(io);
    new_line(io);
    new_line(io);
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Loading                                     = ");
    put(io,loading,15);
    put(io," |");
    new_line(io);

    put(io,"| Minor Iterations                            = ");
    put(io,minor,15);
    put(io," |");
    new_line(io);

    put(io,"| Total Potential Energy                      = ");
    put(io,f,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Allowable Maximum Residue                   = ");
    put(io,maxnorm,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Maximum Residue (Force)                     = ");
    put(io,maxr1,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Maximum Residue (Moment)                    = ");
    put(io,maxr2,2,8,3);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- support reactions
    new_line(io);
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    put(io," Node             Force X             Force Y             Force Z");
    new_line(io);

    put(io,"                 Moment X            Moment Y            Moment Z");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      last1 := 0;
      for axis in 1 .. 3 loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last1 := axis;
        end if;
      end loop;

      last2 := 0;
      for axis in 4 .. lastaxis(glob) loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last2 := axis;
        end if;
      end loop;

      if last1 /= 0 or last2 /= 0 then
        put(io,glob,5);

        for axis in 1 .. last1 loop
          if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
            putchar(io,' ',5);
            put(io,force(glob,axis),2,8,3);
          else
            putchar(io,' ',20);
          end if;
        end loop;

        new_line(io);

        if last2 /= 0 then
          putchar(io,' ',5);
          for axis in 4 .. last2 loop
            if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
              putchar(io,' ',5);
              put(io,force(glob,axis),2,8,3);
            else
              putchar(io,' ',20);
            end if;
          end loop;
          new_line(io);
        end if;
      end if;
    end loop;

    close(io);
  end putreact;

  procedure autocad_1 is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    infinity : constant long_float := 1.0E+20;

    lo,le,xc,yc,zc   : long_float;
    glob,g,ming,maxg : integer;
    io               : file_type;

  begin
    create(io, out_file, "autocad.scr");

    -- set running object snap mode to NONE
    put(io,"osmode 0");
    new_line(io);

    -- cable elements
    if cable'last > 0 then
      put(io,"layer make cable ");
      new_line(io);

      for e in cable'range loop
        put(io,"3dpoly ");
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        new_line(io);
      end loop;
    end if;

    -- membrane elements
    if membr'last > 0 then
      put(io,"layer make membrane ");
      new_line(io);

      for e in membr'range loop
        put(io,"3dface ");
        for loc in 1 .. 3 loop
          glob := membr(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        put(io," ");
        new_line(io);
      end loop;
    end if;

    -- frame elements
    if frame'last > 0 then
      -- group numbering range
      ming := frame'last + 1;
      maxg := 0;

      for e in frame'range loop
        g := frame(e).group;

        if g < ming then
          ming := g;
        end if;

        if g > maxg then
          maxg := g;
        end if;
      end loop;

      for g in ming .. maxg loop
        put(io,"layer make frame_");
        put(io,g,0);
        put(io," ");
        new_line(io);

        for e in frame'range loop
          if frame(e).group = g then
            put(io,"3dpoly ");
            for loc in 1 .. 2 loop
              glob := frame(e).glob(loc);
              put(io,xo(glob),1,8,3);
              put(io,",");
              put(io,yo(glob),1,8,3);
              put(io,",");
              put(io,zo(glob),1,8,3);
              put(io," ");
            end loop;
            new_line(io);
          end if;
        end loop;
      end loop;
    end if;

    -- text size
    lo := infinity;

    -- cable elements
    if cable'last > 0 then
      le := 0.0;
      for e in cable'range loop
        le := le + cable(e).length;
      end loop;

      le := le / long_float(cable'last);

      if le < lo then
        lo := le;
      end if;
    end if;

    -- membrane elements
    if membr'last > 0 then
      le := 0.0;
      for e in membr'range loop
        le := le + membr(e).area;
      end loop;

      le := 1.519671 * sqrt(le / long_float(membr'last));

      if le < lo then
        lo := le;
      end if;
    end if;

    -- frame elements
    if frame'last > 0 then
      le := 0.0;
      for e in frame'range loop
        le := le + frame(e).length;
      end loop;

      le := le / long_float(frame'last);

      if le < lo then
        lo := le;
      end if;
    end if;

    -- text style
    put(io,"textstyle Standard");
    new_line(io);

    -- text size
    put(io,"textsize ");
    put(io,(0.1 * lo),1,8,3);
    new_line(io);

    -- node numbers
    put(io,"layer make node_numbers ");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,"text BL ");

      put(io,xo(glob),1,8,3);
      put(io,",");
      put(io,yo(glob),1,8,3);
      put(io,",");
      put(io,zo(glob),1,8,3);

      put(io,"   ");
      put(io,glob,0);
      put(io," ");

      new_line(io);
    end loop;

    -- cable numbers
    if cable'last > 0 then
      put(io,"layer make cable_numbers ");
      new_line(io);

      for e in cable'range loop
        xc := 0.0;
        yc := 0.0;
        zc := 0.0;
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          xc := xc + xo(glob);
          yc := yc + yo(glob);
          zc := zc + zo(glob);
        end loop;
        xc := xc / 2.0;
        yc := yc / 2.0;
        zc := zc / 2.0;

        put(io,"text MC ");

        put(io,xc,1,8,3);
        put(io,",");
        put(io,yc,1,8,3);
        put(io,",");
        put(io,zc,1,8,3);

        put(io,"   ");
        put(io,e,0);
        put(io," ");

        new_line(io);
      end loop;
    end if;

    -- membrane numbers
    if membr'last > 0 then
      put(io,"layer make membrane_numbers ");
      new_line(io);

      for e in membr'range loop
        xc := 0.0;
        yc := 0.0;
        zc := 0.0;
        for loc in 1 .. 3 loop
          glob := membr(e).glob(loc);
          xc := xc + xo(glob);
          yc := yc + yo(glob);
          zc := zc + zo(glob);
        end loop;
        xc := xc / 3.0;
        yc := yc / 3.0;
        zc := zc / 3.0;

        put(io,"text MC ");

        put(io,xc,1,8,3);
        put(io,",");
        put(io,yc,1,8,3);
        put(io,",");
        put(io,zc,1,8,3);

        put(io,"   ");
        put(io,e,0);
        put(io," ");

        new_line(io);
      end loop;
    end if;

    -- frame numbers
    if frame'last > 0 then
      -- group numbering range
      ming := frame'last + 1;
      maxg := 0;

      for e in frame'range loop
        g := frame(e).group;

        if g < ming then
          ming := g;
        end if;

        if g > maxg then
          maxg := g;
        end if;
      end loop;

      for g in ming .. maxg loop
        put(io,"layer make frame_");
        put(io,g,0);
        put(io,"_numbers ");
        new_line(io);

        for e in frame'range loop
          if frame(e).group = g then
            xc := 0.0;
            yc := 0.0;
            zc := 0.0;
            for loc in 1 .. 2 loop
              glob := frame(e).glob(loc);
              xc := xc + xo(glob);
              yc := yc + yo(glob);
              zc := zc + zo(glob);
            end loop;
            xc := xc / 2.0;
            yc := yc / 2.0;
            zc := zc / 2.0;

            put(io,"text MC ");

            put(io,xc,1,8,3);
            put(io,",");
            put(io,yc,1,8,3);
            put(io,",");
            put(io,zc,1,8,3);

            put(io,"   ");
            put(io,e,0);
            put(io," ");

            new_line(io);
          end if;
        end loop;
      end loop;
    end if;

    close(io);
  end autocad_1;

  procedure autocad_2 (x : in terms.fmat1_ptr) is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    glob : integer;
    v    : fmat3x1;
    io   : file_type;

  begin
    open(io, append_file, "autocad.scr");

    put(io,"layer make loading_");
    put(io,loading,0);
    put(io," ");
    new_line(io);

    -- cable elements
    if cable'last > 0 then
      for e in cable'range loop
        if not cable(e).mark then
          put(io,"3dpoly ");
          for loc in 1 .. 2 loop
            glob := cable(e).glob(loc);
            vertex(glob,v);
            put(io,v(1),1,8,3);
            put(io,",");
            put(io,v(2),1,8,3);
            put(io,",");
            put(io,v(3),1,8,3);
            put(io," ");
          end loop;
          new_line(io);
        end if;
      end loop;

      put(io,"-color");
      new_line(io);
      put(io,"red");
      new_line(io);

      for e in cable'range loop
        if cable(e).mark then
          put(io,"3dpoly ");
          for loc in 1 .. 2 loop
            glob := cable(e).glob(loc);
            vertex(glob,v);
            put(io,v(1),1,8,3);
            put(io,",");
            put(io,v(2),1,8,3);
            put(io,",");
            put(io,v(3),1,8,3);
            put(io," ");
          end loop;
          new_line(io);
        end if;
      end loop;

      put(io,"-color");
      new_line(io);
      put(io,"bylayer");
      new_line(io);
    end if;

    -- membrane elements
    if membr'last > 0 then
      for e in membr'range loop
        if not membr(e).mark then
          put(io,"3dface ");
          for loc in 1 .. 3 loop
            glob := membr(e).glob(loc);
            vertex(glob,v);
            put(io,v(1),1,8,3);
            put(io,",");
            put(io,v(2),1,8,3);
            put(io,",");
            put(io,v(3),1,8,3);
            put(io," ");
          end loop;
          put(io," ");
          new_line(io);
        end if;
      end loop;

      put(io,"-color");
      new_line(io);
      put(io,"red");
      new_line(io);

      for e in membr'range loop
        if membr(e).mark then
          put(io,"3dface ");
          for loc in 1 .. 3 loop
            glob := membr(e).glob(loc);
            vertex(glob,v);
            put(io,v(1),1,8,3);
            put(io,",");
            put(io,v(2),1,8,3);
            put(io,",");
            put(io,v(3),1,8,3);
            put(io," ");
          end loop;
          put(io," ");
          new_line(io);
        end if;
      end loop;

      put(io,"-color");
      new_line(io);
      put(io,"bylayer");
      new_line(io);
    end if;

    -- frame elements
    for e in frame'range loop
      put(io,"3dpoly ");
      for loc in 1 .. 2 loop
        glob := frame(e).glob(loc);
        vertex(glob,v);
        put(io,v(1),1,8,3);
        put(io,",");
        put(io,v(2),1,8,3);
        put(io,",");
        put(io,v(3),1,8,3);
        put(io," ");
      end loop;
      new_line(io);
    end loop;

    close(io);
  end autocad_2;

  procedure autocad_3 is

    use
      text_io;

    g,ming,maxg : integer;
    io          : file_type;

  begin
    open(io, append_file, "autocad.scr");

    put(io,"layer set 0 ");
    new_line(io);

    -- node numbers
    put(io,"layer off node_numbers ");
    new_line(io);

    -- cable numbers
    if cable'last > 0 then
      put(io,"layer off cable_numbers ");
      new_line(io);
    end if;

    -- membrane numbers
    if membr'last > 0 then
      put(io,"layer off membrane_numbers ");
      new_line(io);
    end if;

    -- frame numbers
    if frame'last > 0 then
      -- group numbering range
      ming := frame'last + 1;
      maxg := 0;

      for e in frame'range loop
        g := frame(e).group;

        if g < ming then
          ming := g;
        end if;

        if g > maxg then
          maxg := g;
        end if;
      end loop;

      for g in ming .. maxg loop
        put(io,"layer off frame_");
        put(io,g,0);
        put(io,"_numbers ");
        new_line(io);
      end loop;
    end if;

    for l in 1 .. nl loop
      put(io,"layer off loading_");
      put(io,l,0);
      put(io," ");
      new_line(io);
    end loop;

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad_3;

begin
  getmaterial;

  getstructure;

  autocad_1;

  open(io, in_file, "loading.txt");

  -- number of loadings
  scan(io);
  get(io,nl);
  skip_line(io);

  clock1 := clock;

  for l in 1 .. nl loop
    getloading(io,loading,x);

    qn.lbfgs(maxlist,n,maxminor,maxls,
             minor,
             maxnorm,lstol,
             f,
             x,g,
             optimal);

    if optimal then
      put("Optimal");
    else
      put("NOT Optimal");
    end if;
    new_line;

    put("Minor = ");
    put(minor,5);
    new_line;

    finish(x);

    putcable(x);

    putmembr(x);

    putreact(minor,f,maxnorm,x);

    putdispl(x);

    putcoord(x);

    putframe(x);

    putspring(x);

    autocad_2(x);
  end loop;

  autocad_3;

  clock2 := clock;

  new_line;
  put("Elapsed time = ");
  put(integer(clock2 - clock1));
  put(" seconds");
  new_line;

  close(io);
end lights;
